/*		 IIT-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: itifgReg.c,v 7.0 2000/10/25 09:07:57 cvs Exp $
 *
 */

/*
 * Defines:
 */

/*
 * Includes:
 */

#include "itifgOS.h"	/* os specific stuff */
#include "itifgExt.h"	/* offical driver interface */
#include "itifgReg.h"	/* register definitions ICR Regs */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */
#include "pcdigReg.h"	/* register definitions PC-DIG */
#include "itifgIntD.h"	/* driver internals */
#include "itifgIntF.h"	/* internal functions */

/*
 * Globals:
 */

#if (DEBUG_ITIFG & 128)
const char *icr_reg[] = { "mbox1 ",
			  "mbox2 ",
			  "mbox3 ",
			  "mbox4 ",
			  "error ",
			  "error ",
			  "error ",
			  "error ",
			  "error ",
			  "bmdst ",
			  "bmxc  ",
			  "error ",
			  "error ",
			  "error ",
			  "intctl",
			  "bmctl "
			};

#endif

/*
 * Prototypes:
 */

u_long icr_read_mm (itifg_t*, u_char, u_long, u_long, u_long);
u_char icr_read_ms (itifg_t*, u_char, u_long, u_char);
int icr_write_mm (itifg_t*, u_char, u_long, u_long, u_long, u_long);
int icr_write_ms (itifg_t*, u_char, u_long, u_char, u_char);

/*
 * General Read Handler for ITI-FG Interface Control Registers
 */

/* 32 bit register access with mask and range checking */

u_long
icr_read_mm (iti, offset, mask, min, max)
     
  itifg_t *iti;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_long mask;		/* read mask for register */
  u_long min;		/* minimal value for range checking */
  u_long max;		/* maximal value for range checking */

{
  u_long tmp_val = 0;

  if (!iti->regs_io)
    tmp_val = iti_readl (iti->v_icr_base + offset /
			 sizeof(u_long));
  else
    tmp_val = iti_inl (iti->io_icr_base + offset);
  
  tmp_val &= mask;

#if(0)
  if (tmp_val & ~mask) iti->errno = -EITI_MASK;

  if (tmp_val < min || tmp_val > max) iti->errno = -EITI_RANGE;
#endif

  if (!iti->regs_io)
    ITI_PRINT7A("icr-readl %s, %#010lX &(%#010lX) >%#010lX <%#010lX\n",
		icr_reg[offset / sizeof(u_long)], tmp_val, mask, min, max);
  else
    ITI_PRINT7A("icr-inl %s, %#010lX &(%#010lX) >%#010lX <%#010lX\n",
		icr_reg[offset / sizeof(u_long)], tmp_val, mask, min, max);

  return (tmp_val);
}

/* 32 bit register 1 (or more) bit access */

u_char
icr_read_ms (iti, offset, mask, shift)

  itifg_t *iti;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_long mask;		/* read mask for register */
  u_char shift;		/* number of bits to shift right */

{
  u_long tmp_val = 0;

  if (!iti->regs_io)
    tmp_val = iti_readl (iti->v_icr_base + offset /
			 sizeof(u_long));
  else
    tmp_val = iti_inl (iti->io_icr_base + offset);

  tmp_val &= mask;
  tmp_val >>= shift;

  if (!iti->regs_io)
    ITI_PRINT7A("icr-readl %s, %#1lX &(%#010lX) >>%2d\n",
		icr_reg[offset / sizeof(u_long)], tmp_val, mask, shift);
  else
    ITI_PRINT7A("icr-inl %s, %#1lX &(%#010lX) >>%2d\n",
		icr_reg[offset / sizeof(u_long)], tmp_val, mask, shift);

  return ((u_char)tmp_val);
}
  
/*
 * General Write Handler for Interface Control Registers
 */

/* 32 bit register access with mask and range checking */

int
icr_write_mm (iti, offset, mask, min, max, value)

  itifg_t *iti;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_long mask;		/* read mask for register */
  u_long min;		/* minimal value for range checking */
  u_long max;		/* maximal value for range checking */
  u_long value;		/* value to set to */

{
  u_long tmp_val = 0;

  if (value & ~mask) {
    iti->errno = -EITI_MASK;
    return (ERROR);
  }

  if (value < min || value > max) {
    iti->errno = -EITI_RANGE;  
    return (ERROR);
  }

  tmp_val = value & mask;

  if (!iti->regs_io)
    ITI_PRINT7A("icr-writel %s, %#010lX &(%#010lX) >%#010lX <%#010lX\n",
		icr_reg[offset / sizeof(u_long)], tmp_val, mask, min, max);
  else
    ITI_PRINT7A("icr-outl %s, %#010lX &(%#010lX) >%#010lX <%#010lX\n",
		icr_reg[offset / sizeof(u_long)], tmp_val, mask, min, max);
  
  if (!iti->regs_io)
    iti_writel (tmp_val, iti->v_icr_base + offset /
		sizeof(u_long));
  else
    iti_outl (tmp_val, iti->io_icr_base + offset);

  return (OK);
}

/* 32 bit register 1 (or more) bit access */

int
icr_write_ms (iti, offset, mask, shift, value)

  itifg_t *iti;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_long mask;		/* read mask for register */
  u_char shift;		/* number of bits to shift right */
  u_char value;		/* value to set to */  

{
  u_long read_val = 0;
  u_long tmp_val = 0;

  tmp_val = (u_long)value << shift;

  if (tmp_val & ~mask) {
    iti->errno = -EITI_MASK;
    return (ERROR);
  }

  /* this have to be atomar */
  /* iti_splhigh (iti); */

  if (!iti->regs_io)
    read_val = iti_readl (iti->v_icr_base + offset /
			  sizeof(u_long));
  else
    read_val = iti_inl (iti->io_icr_base + offset);
  
  read_val &= ~mask;
  read_val |= tmp_val;

  if (!iti->regs_io)
    ITI_PRINT7A("icr-writel %s, %#010lX &(%#010lX) <<%2d\n",
		icr_reg[offset / sizeof(u_long)], tmp_val, mask, shift);
  else
    ITI_PRINT7A("icr-outl %s, %#010lX &(%#010lX) <<%2d\n",
		icr_reg[offset / sizeof(u_long)], tmp_val, mask, shift);

  if (!iti->regs_io)
    iti_writel (read_val, iti->v_icr_base + offset /
		sizeof(u_long));
  else
    iti_outl (read_val, iti->io_icr_base + offset);

  /* iti_spllow (iti); */ /* end atomar */

  return (OK);
}
  
/*
 * Special Read Handler for IC-PCI Interface Control Registers
 */

u_long
icr_get_mbox1 (itifg_t *iti) {

  return (icr_read_mm (iti,
		       ICR_MBOX1_OFFSET,
		       ICR_MBOX_MASK,
		       ICR_MBOX_MIN,
		       ICR_MBOX_MAX));
}

u_long
icr_get_mbox2 (itifg_t *iti) {

  return (icr_read_mm (iti,
		       ICR_MBOX2_OFFSET,
		       ICR_MBOX_MASK,
		       ICR_MBOX_MIN,
		       ICR_MBOX_MAX));
}

u_long
icr_get_mbox3 (itifg_t *iti) {

  return (icr_read_mm (iti,
		       ICR_MBOX3_OFFSET,
		       ICR_MBOX_MASK,
		       ICR_MBOX_MIN,
		       ICR_MBOX_MAX));
}

u_long
icr_get_mbox4 (itifg_t *iti) {

  return (icr_read_mm (iti,
		       ICR_MBOX4_OFFSET,
		       ICR_MBOX_MASK,
		       ICR_MBOX_MIN,
		       ICR_MBOX_MAX));
}

u_long
icr_get_bmdst (itifg_t *iti) {

  return (icr_read_mm (iti,
		       ICR_BMDST_OFFSET,
		       ICR_BMDST_MASK,
		       ICR_BMDST_MIN,
		       ICR_BMDST_MAX));
}

u_long
icr_get_bmxc (itifg_t *iti) {
  
  if (iti->big_mem) {
    return (icr_read_mm (iti,
			 ICR_BMXC_OFFSET,
			 (ICR_BMXC_MASK << 1) + 1,
			 ICR_BMXC_MIN,
			 (ICR_BMXC_MAX << 1) + 1));
  } else {
    return (icr_read_mm (iti,
			 ICR_BMXC_OFFSET,
			 ICR_BMXC_MASK,
			 ICR_BMXC_MIN,
			 ICR_BMXC_MAX));
  }
}

u_char
icr_get_inten (itifg_t *iti) {

    return (icr_read_ms (iti,
			 ICR_INTCTL_OFFSET,
			 ICR_INTEN_MASK,
			 ICR_INTEN_SHIFT));
}

u_char
icr_get_binten (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_INTCTL_OFFSET,
		       ICR_INTEN_MASK,
		       ICR_INTEN_SHIFT));
}

/* old, use xcr_test_irq */
u_char
icr_get_intst (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_INTCTL_OFFSET,
		       ICR_INTST_MASK,
		       ICR_INTST_SHIFT));
}

/* old, use xcr_test_irq */
u_char
icr_get_bintst (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_INTCTL_OFFSET,
		       ICR_BINTST_MASK,
		       ICR_BINTST_SHIFT));
}

/* old, use xcr_test_irq */
u_char
icr_get_maint (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_INTCTL_OFFSET,
		       ICR_MAINT_MASK,
		       ICR_MAINT_SHIFT));
}

/* old, use xcr_test_irq */
u_char
icr_get_taint (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_INTCTL_OFFSET,
		       ICR_TAINT_MASK,
		       ICR_TAINT_SHIFT));
}

u_char
icr_get_fifofl (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_BMCTL_OFFSET,
		       ICR_FIFOFL_MASK,
		       ICR_FIFOFL_SHIFT));
}

u_char
icr_get_fifo4p (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_BMCTL_OFFSET,
		       ICR_FIFO4P_MASK,
		       ICR_FIFO4P_SHIFT));
}

u_char
icr_get_fifoem (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_BMCTL_OFFSET,
		       ICR_FIFOEM_MASK,
		       ICR_FIFOEM_SHIFT));
}

u_char
icr_get_bmdone (itifg_t *iti) {

  return (icr_read_ms (iti,
		       ICR_BMCTL_OFFSET,
		       ICR_BMDONE_MASK,
		       ICR_BMDONE_SHIFT));
}

u_char
icr_get_bmreq (itifg_t *iti) {

  if (IS_PCDIG)
    {
      iti->errno = -EITI_NOSUPPORT;
      return ERROR;
    }

  return (icr_read_ms (iti,
		       ICR_BMCTL_OFFSET,
		       ICR_BMREQ_MASK,
		       ICR_BMREQ_SHIFT));
}

/*
 * Special Write Handler for IC-PCI Interface Control Registers
 */

int
icr_set_mbox1 (itifg_t *iti, u_long value) {

  return (icr_write_mm (iti,
			ICR_MBOX1_OFFSET,
			ICR_MBOX_MASK,
			ICR_MBOX_MIN,
			ICR_MBOX_MAX,
			value));
}

int
icr_set_mbox2 (itifg_t *iti, u_long value) {

  return (icr_write_mm (iti,
			ICR_MBOX2_OFFSET,
			ICR_MBOX_MASK,
			ICR_MBOX_MIN,
			ICR_MBOX_MAX,
			value));
}

int
icr_set_mbox3 (itifg_t *iti, u_long value) {

  return (icr_write_mm (iti,
			ICR_MBOX3_OFFSET,
			ICR_MBOX_MASK,
			ICR_MBOX_MIN,
			ICR_MBOX_MAX,
			value));
}

int
icr_set_mbox4 (itifg_t *iti, u_long value) {

  return (icr_write_mm (iti,
			ICR_MBOX4_OFFSET,
			ICR_MBOX_MASK,
			ICR_MBOX_MIN,
			ICR_MBOX_MAX,
			value));
}

int
icr_set_bmdst (itifg_t *iti, u_long value) {

  if (IS_PCDIG)
    {
      iti->errno = -EITI_NOSUPPORT;
      return ERROR;
    }

  return (icr_write_mm (iti,
			ICR_BMDST_OFFSET,
			ICR_BMDST_MASK,
			ICR_BMDST_MIN,
			ICR_BMDST_MAX,
			value));
}

int
icr_set_bmxc (itifg_t *iti, u_long value) {

  if (IS_PCDIG)
    {
      iti->errno = -EITI_NOSUPPORT;
      return ERROR;
    }

  if (iti->big_mem) {
    return (icr_write_mm (iti,
			  ICR_BMXC_OFFSET,
			  (ICR_BMXC_MASK << 1) + 1,
			  ICR_BMXC_MIN,
			  (ICR_BMXC_MAX << 1) + 1,
			  value));
  } else {
    return (icr_write_mm (iti,
			  ICR_BMXC_OFFSET,
			  ICR_BMXC_MASK,
			  ICR_BMXC_MIN,
			  ICR_BMXC_MAX,
			  value));
  }
}

/* old, use iti_[en|dis]able_irq */
int
icr_set_inten (itifg_t *iti, u_char value) {

  return (icr_write_ms (iti,
			ICR_INTCTL_OFFSET,
			ICR_INTEN_MASK,
			ICR_INTEN_SHIFT,
			value));
	    
}

/* old, use iti_[en|dis]able_irq */
int
icr_set_binten (itifg_t *iti, u_char value) {

  return (icr_write_ms (iti,
			ICR_INTCTL_OFFSET,
			ICR_BINTEN_MASK,
			ICR_BINTEN_SHIFT,
			value));
}

int
icr_set_bmreq (itifg_t *iti, u_char value) {

  if (IS_PCDIG)
    {
      iti->errno = -EITI_NOSUPPORT;
      return ERROR;
    }

  if (value == ICR_ENABLE) {
    return (icr_write_mm (iti,
			  ICR_BMCTL_OFFSET,
			  ICR_NO_MASK,
			  ICR_NO_MIN,
			  ICR_NO_MAX,
			  ICR_STRT_TRANS));
  } else {
    return (icr_write_mm (iti,
			  ICR_BMCTL_OFFSET,
			  ICR_NO_MASK,
			  ICR_NO_MIN,
			  ICR_NO_MAX,
			  ICR_STOP_TRANS));
  }
}

/*
 * Special Clear Handler for IC-PCI Interface Control Registers
 */

int
icr_clear_fifoclr (itifg_t *iti) {

    return (icr_write_ms (iti,
			  ICR_BMCTL_OFFSET,
			  ICR_FIFOCLR_MASK,
			  ICR_FIFOCLR_SHIFT,
			  ICR_ENABLE));
}

/*
 * Special Initialization Handler for Interface Control Registers
 */

int
icr_init_intctl (itifg_t *iti) {

    return (icr_write_mm (iti,
			  ICR_INTCTL_OFFSET,
			  ICR_NO_MASK,
			  ICR_NO_MIN,
			  ICR_NO_MAX,
			  ICR_INTCTL_INIT));
}
		
int
icr_init_bmctl (itifg_t *iti) {

  return (icr_write_mm (iti,
			ICR_BMCTL_OFFSET,
			ICR_NO_MASK,
			ICR_NO_MIN,
			ICR_NO_MAX,
			ICR_BMCTL_INIT));
}

/*
 * Special Interrupt clear function (called by iti_intr/iti_en|disable_irq)
 */

void
xcr_clear_irqs (itifg_t *iti)
{

  u_short cr_additional = 0;
  u_short cr_standard = 0;

  extern u_char
    cr_read_ms (itifg_t*, u_char, u_char, u_short);
  extern u_short
    cr_read_mm (itifg_t*, u_char, u_short, u_short, u_short);
  extern int
    cr_write_ms (itifg_t*, u_char, u_short, u_char, u_char);
  extern int
    cr_write_mm (itifg_t*, u_char, u_short, u_short, u_short, u_short);

  switch (DEVICE_ID)
    {
    case ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI:

      /* mark and clear standard IC-PCI interrupts */
      cr_standard = cr_read_mm (iti,
				CR_INTCON_OFFSET,
				CR_NO_MASK,
				CR_NO_MIN,
				CR_NO_MAX);
      cr_write_mm (iti,
		   CR_INTCON_OFFSET,
		   CR_NO_MASK,
		   CR_NO_MIN,
		   CR_NO_MAX,
		   CR_INTCON_INIT);

      /* mark and clear additional IC-PCI REV-B and PCVision interrupts */
      if (IS_REV_B)
	{
	  if (cr_read_ms (iti,
			  CR_PORTCON_OFFSET,
			  CR_INPIEN_MASK,
			  CR_INPIEN_SHIFT))
	    cr_additional |= CR_INPINT;
	  cr_write_ms (iti,
		       CR_PORTCON_OFFSET,
		       CR_INPIEN_MASK,
		       CR_INPIEN_SHIFT,
		       CR_DISABLE);
	}
      
      /* clear additional PCVision interrupts */
      if (IS_AMPV)
	iti_writew (iti_readw (&iti->pv_reg->con2.all) | PV_IRQ_TEST_MASK,
		    &iti->pv_reg->con2.all);
      break;

    case ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG:
      if (FPGA_OK)
	{
	  /* clear interrupts from source 0 */
	  iti_writel (PCD_IRQ0_TEST_MASK, &iti->pcd_ca_reg->intstat0.all);
	  /* clear interrupts from source 1 */
	  iti_writel (PCD_IRQ1_TEST_MASK, &iti->pcd_bi_reg->intstat1.all);
	}
      break;
    }

  /* clear main AMCC (PCI-Matchmaker) interrupts */
  icr_write_mm (iti,
		ICR_INTCTL_OFFSET,
		ICR_NO_MASK,
		ICR_NO_MIN,
		ICR_NO_MAX,
		icr_read_mm (iti,
			     ICR_INTCTL_OFFSET,
			     ICR_NO_MASK,
			     ICR_NO_MIN,
			     ICR_NO_MAX) | ICR_INTCTL_INIT);

  /* reenable standard IC-PCI interrupts */
  if (IS_ICPCI)
    cr_write_mm (iti,
		 CR_INTCON_OFFSET,
		 CR_NO_MASK,
		 CR_NO_MIN,
		 CR_NO_MAX,
		 cr_standard);

  /* reenable additional IC-PCI REV-B and PCVision interrupts */
  if (IS_ICPCI)
    if (IS_REV_B)
      if (cr_additional & CR_INPINT)
	cr_write_ms (iti,
		     CR_PORTCON_OFFSET,
		     CR_INPIEN_MASK,
		     CR_INPIEN_SHIFT,
		     CR_ENABLE);
}

/*
 * Special Interrupt test function (called by iti_intr)
 */

void
xcr_test_irqs (itifg_t *iti, u_long *irqs)
{

  u_short pv_additional = 0;
  u_short cr_additional = 0;
  u_short cr_standard = 0;
  u_long icr_standard = 0UL;
  u_long pcd_source0 = 0UL;
  u_long pcd_source1 = 0UL;

  extern u_char
    cr_read_ms (itifg_t*, u_char, u_char, u_short);
  extern u_short
    cr_read_mm (itifg_t*, u_char, u_short, u_short, u_short);

  /* test main AMCC (PCI-Matchmaker) interrupts */  
  icr_standard = icr_read_mm (iti,
			      ICR_INTCTL_OFFSET,
			      ICR_NO_MASK,
			      ICR_NO_MIN,
			      ICR_NO_MAX);
  icr_standard &= ICR_IRQ_TEST_MASK;

  *irqs |= icr_standard;

  /* only test the other irq's, if the main ones are detected */
  if (*irqs)
    {
      switch (DEVICE_ID)
	{
	case ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI:
	  
	  /* test standard IC-PCI interrupts */  
	  cr_standard = cr_read_mm (iti,
				    CR_INTCON_OFFSET,
				    CR_NO_MASK,
				    CR_NO_MIN,
				    CR_NO_MAX);
	  cr_standard &= CR_IRQ_TEST_MASK;

	  /* test additional IC-PCI REV-B and PCVision interrupts */
	  if (IS_REV_B)
	    if (cr_read_ms (iti,
			    CR_PORTCON_OFFSET,
			    CR_INPINT_MASK,
			    CR_INPINT_SHIFT))
	      cr_additional |= CR_INPINT;
	  
	  /* test additional PCVision interrupts */
	  if (IS_AMPV)
	    {
	      pv_additional = iti_readw (&iti->pv_reg->con2.all);
	      pv_additional &= PV_IRQ_TEST_MASK;
	      pv_additional >>= PV_IRQ_TEST_SHIFT;
	    }
	  
      *irqs |= cr_standard | cr_additional | pv_additional;
      break;
      
	case ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG:
	  if (FPGA_OK)
	    {
	      /* test interrupts from source 0 */
	      pcd_source0 = iti_readl (&iti->pcd_ca_reg->intstat0.all);
	      pcd_source0 &= PCD_IRQ0_TEST_MASK;
	      pcd_source0 <<= PCD_IRQ0_TEST_SHIFT;
	      /* test interrupts from source 1 */
	      pcd_source1 = iti_readl (&iti->pcd_bi_reg->intstat1.all);
	      pcd_source1 &= PCD_IRQ1_TEST_MASK;
	      pcd_source1 <<= PCD_IRQ1_TEST_SHIFT;
	      *irqs |= pcd_source0 | pcd_source1;
	    }
	  break;
	}
    }
}

